/*
  
  mj 3.5.2  Copyright (C) 2007 Marco Bisetto < marco [at] folgorante.net >
  This program comes with ABSOLUTELY NO WARRANTY; 
  This is free software, and you are welcome to redistribute it
  under certain conditions;

  Released under the terms of GNU GPLv3.

  Computes Mandelbrot and Julia sets.
  Gets data from commandline.
  Returns a word matrix with binary header in stdout.
  The output matrix can be transformed in ppm image file using mj2ppm

  Calcola insiemi di Mandelbrot e di Julia.
  Accetta i dati dalla riga di comando.
  Restituisce una matrice di word con un intestazione binaria in stdout.
  
  La matrice puo` essere trasformata in file immagine ppm con mj2ppm.
  
*/

#include <stdio.h>
#include <stdlib.h>
#include "mj-3.0.h"

static inline unsigned int punto(struct mjpt ss)
{
  unsigned int n;
  
  asm ("\tfninit\n");

  asm (
       "fld %%st(1)\n\t"
       "fmul %%st(0)\n\t"
       "fld %%st(1)\n\t"
       "fmul %%st(0)\n\t"
       "movl $0,%%ecx\n"
     "ritor:\t"
       "fsubrp %%st,%%st(1)\n\t"
       "fadd %%st(3)\n\t"
       "fxch %%st(2)\n\t"
       "fmulp %%st,%%st(1)\n\t"
       "fadd %%st(0)\n\t"
       "fadd %%st(3)\n\t"
       "fld %%st(1)\n\t"
       "fmul %%st(0)\n\t"
       "fld %%st(1)\n\t"
       "fmul %%st(0)\n\t"
       "fld %%st(0)\n\t"
       "fadd %%st(2)\n\t"
       "fcomp %%st(7)\n\t"
       "fnstsw %%ax\n\t"
       "and $0x41, %%ah\n\t"
       "jz esci\n\t"
       "incl %%ecx\n\t"
       "cmpl %%ebx,%%ecx\n\t"
       "jb ritor\n"
     "esci:\n"  
       : "=c" (n)
       : "b" (ss.ma), "f" (ss.mam2), "f" (ss.yc), "f" (ss.xc),
         "f" (ss.xi), "f" (ss.yi)
       : "%eax"
       );

  asm ("\tfinit\n");
  
  return n;
}

void mandel(struct mjim t)
{
    double x1,y1,ps=t.l0/(double)t.xn;
    unsigned int a,b;
    unsigned short n;
    struct mjpt s;
  
    s.ma=t.ma;
    s.mam2=t.mam2;

    y1=t.y0;
    for (b=0; b<t.yn; b++) {
	s.yi=y1;
	s.yc=y1;
	x1=t.x0;
	for (a=0; a<t.xn; a++) {
	    s.xi=x1;
	    s.xc=x1;
	    n=(unsigned short)punto(s);
	    fwrite(&n,sizeof(n),1,stdout);
	    x1+=ps;
	}
	y1-=ps;
    }
}

void julia(struct mjim t)
{
    double x1,y1,ps=t.l0/(double)t.xn;
    unsigned int a,b;
    unsigned short n;
    struct mjpt s;
  
    s.ma=t.ma;
    s.mam2=t.mam2;
    s.xc=t.xj;
    s.yc=t.yj;

    y1=t.y0;
    for (b=0; b<t.yn; b++) {
	s.yi=y1;
	x1=t.x0;
	for (a=0; a<t.xn; a++) {
	    s.xi=x1;
	    n=(unsigned short)punto(s);
	    fwrite(&n,sizeof(n),1,stdout);
	    x1+=ps;
	}
	y1-=ps;
    }
}

int main(int argc, char * argv[])
{
    struct mjim t;

    switch (argc) {
    case 8:
	t.tipo=MAN;
	t.xn=atoi(argv[1]);
	t.yn=atoi(argv[2]);    
	t.ma=atoi(argv[3]);
	t.mam2=atof(argv[4]);
	t.x0=atof(argv[5]);
	t.y0=atof(argv[6]);
	t.l0=atof(argv[7]);
	fwrite(&t,sizeof(t),1,stdout);
	fprintf(stderr,"\n Mandelbrot\n xn=%u\n yn=%u\n ma=%u\n mam2=%f\n"
		" x0=%.17e\n y0=%.17e\n l0=%.17e\n",
		t.xn, t.yn, t.ma, t.mam2, t.x0, t.y0, t.l0);
	mandel(t);
	break;
    case 10:
	t.tipo=JUL;
	t.xn=atoi(argv[1]);
	t.yn=atoi(argv[2]);    
	t.ma=atoi(argv[3]);
	t.mam2=atof(argv[4]);
	t.x0=atof(argv[5]);
	t.y0=atof(argv[6]);
	t.l0=atof(argv[7]);
	t.xj=atof(argv[8]);
	t.yj=atof(argv[9]);
	fwrite(&t,sizeof(t),1,stdout);
	fprintf(stderr,"\n Julia\n xn=%u\n yn=%u\n ma=%u\n mam2=%f\n"
		" x0=%.17e\n y0=%.17e\n l0=%.17e\n xj=%.17e\n yj=%.17e\n",
		t.xn, t.yn, t.ma, t.mam2, t.x0, t.y0, t.l0, t.xj, t.yj);
	julia(t);
	break;
    default:
	fprintf(stderr,"\n mj (int)xn (int)yn (int)ma (double)mam2 (double)x0"
		" (double)y0\n (double)l0 [(double)xj (double)yj]\n\n");
    }
}
